__version__ = "0.8.1"


TYPE = {
    0:  'bios',
    1:  'system',
    2:  'base board',
    3:  'chassis',
    4:  'processor',
    7:  'cache',
    8:  'port connector',
    9:  'system slot',
    10: 'on board device',
    11: 'OEM strings',
    #13: 'bios language',
    15: 'system event log',
    16: 'physical memory array',
    17: 'memory device',
    19: 'memory array mapped address',
    24: 'hardware security',
    25: 'system power controls',
    27: 'cooling device',
    32: 'system boot',
    41: 'onboard device',
    }


def parse_dmi(content):
    """
    Parse the whole dmidecode output.
    Returns a list of tuples of (type int, value dict).
    """
    info = []
    lines = iter(content.strip().splitlines())
    while True:
        try:
            line = lines.next()
        except StopIteration:
            break

        if line.startswith('Handle 0x'):
            typ = int(line.split(',', 2)[1].strip()[len('DMI type'):])
            if typ in TYPE:
                info.append((TYPE[typ], _parse_handle_section(lines)))
    return info


def _parse_handle_section(lines):
    """
    Parse a section of dmidecode output

    * 1st line contains address, type and size
    * 2nd line is title
    * line started with one tab is one option and its value
    * line started with two tabs is a member of list
    """
    data = {
        '_title': lines.next().rstrip(),
        }

    for line in lines:
        line = line.rstrip()
        if line.startswith('\t\t'):
            data[k].append(line.lstrip())
        elif line.startswith('\t'):
            k, v = [i.strip() for i in line.lstrip().split(':', 1)]
            if v:
                data[k] = v
            else:
                data[k] = []
        else:
            break

    return data


def profile():
    import os, sys
    if os.isatty(sys.stdin.fileno()):
        content = _get_output()
    else:
        content = sys.stdin.read()

    info = parse_dmi(content)
    _show(info)


def _get_output():
    import subprocess
    output = subprocess.check_output(
        'PATH=$PATH:/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin '
        'sudo dmidecode', shell=True)
    return output


def _show(info):
    def _get(i):
        return [v for j, v in info if j == i]

    system = _get('system')[0]
    print '%s %s (SN: %s, UUID: %s)' % (
        system['Manufacturer'],
        system['Product Name'],
        system['Serial Number'],
        system['UUID'],
        )

    for cpu in _get('processor'):
        print '%s %s %s (Core: %s, Thead: %s)' % (
            cpu['Manufacturer'],
            cpu['Family'],
            cpu['Max Speed'],
            cpu['Core Count'],
            cpu['Thread Count'],
            )

    cnt, total, unit = 0, 0, None
    for mem in _get('memory device'):
        if mem['Size'] == 'No Module Installed':
            continue
        i, unit = mem['Size'].split()
        cnt += 1
        total += int(i)
    print '%d memory stick(s), %d %s in total' % (
        cnt,
        total,
        unit,
        )


if __name__ == '__main__':
    profile()


from canaryd.packages import six


class Change(object):
    def __init__(self, plugin, type_, key, data=None):
        self.plugin = plugin
        self.type = type_
        self.key = key

        self.data = data

    def serialise(self):
        data = None

        if self.type == 'updated':
            data = self.plugin.serialise_update(self.data)

        elif self.type == 'added':
            data = self.plugin.serialise_item(self.data)

        return self.type, self.key, data


def make_diffs(plugin, changes):
    return [
        Change(plugin, *change)
        for change in changes
    ]


def get_state_diff(plugin, plugin_state, previous_plugin_state):
    '''
    Diffs two state dicts and returns a list of changes.

    Changes:
        All come as a tuple of ``(type, key, data, message=None)``.
    '''

    changes = []

    # Look through the previous state to find any items that have been removed
    for key, state in six.iteritems(previous_plugin_state):
        if key not in plugin_state:
            changes.append(Change(plugin, 'deleted', key))

    # Loop the new state
    for key, item in six.iteritems(plugin_state):
        previous_item = previous_plugin_state.get(key)

        # Add anything that doesn't exist
        if not previous_item:
            changes.append(Change(plugin, 'added', key, item))
            continue

        # Plugins customise diff
        if plugin.is_change(key, previous_item, item) is False:
            continue

        # Create the diff, which is a key -> (old, new) values
        state_diff = dict(
            (k, (previous_item.get(k), v))
            for k, v in six.iteritems(item)
            if v != previous_item.get(k)
        )

        if state_diff:
            changes.append(Change(plugin, 'updated', key, state_diff))

    return changes



# canaryd
# File: canaryd/plugin.py
# Desc: functions for handling the various built in state gathering plugins

import re

from glob import glob
from os import path
from subprocess import CalledProcessError, PIPE

from canaryd.packages import six
from canaryd.packages.check_output import check_output
from canaryd.packages.importlib import import_module

from canaryd.log import logger, print_exception # noqa

PLUGINS = []
NAME_TO_PLUGIN = {}


def underscore(name):
    '''
    Transform CamelCase -> snake_case.
    '''

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


class PluginMeta(type):
    def __init__(cls, name, bases, attrs):
        if name != 'Plugin':
            name = underscore(name)

            plugin_instance = cls()
            plugin_instance.name = name

            PLUGINS.append(plugin_instance)
            NAME_TO_PLUGIN[name] = plugin_instance


@six.add_metaclass(PluginMeta)
class Plugin(object):
    spec = None
    command = None
    prepare_command = None
    parent = None

    @staticmethod
    def is_change(key, previous_item, item):
        return True

    def log(self, message):
        message = '[{0}]: {1}'.format(self.name, message)
        logger.debug(message)

    def check_spec_key(self, key, value):
        spec = self.spec[1]

        if key not in spec:
            raise ValueError(
                'Key {0} not found in plugin {1} spec: {2}'.format(
                    key,
                    self.name,
                    value,
                ),
            )

    # Serialisation
    #

    def serialise_value(self, key, value):
        spec = self.spec[1]
        self.check_spec_key(key, value)

        if isinstance(spec[key], set):
            return list(value)

        return value

    def serialise_item(self, item):
        return dict(
            (key, self.serialise_value(key, value))
            for key, value in six.iteritems(item)
        )

    def serialise_state(self, state):
        return dict(
            (key, self.serialise_item(item))
            for key, item in six.iteritems(state)
        )

    def serialise_update(self, changes):
        return dict(
            (key, (self.serialise_value(key, item[0]), self.serialise_value(key, item[1])))
            for key, item in six.iteritems(changes)
        )

    # Unserialisation
    #

    def unserialise_value(self, key, value):
        spec = self.spec[1]
        self.check_spec_key(key, value)

        if isinstance(spec[key], set):
            return set(value)

        return value

    def unserialise_item(self, item):
        return dict(
            (key, self.unserialise_value(key, value))
            for key, value in six.iteritems(item)
        )

    def unserialise_state(self, items):
        return dict(
            (key, self.unserialise_item(item))
            for key, item in six.iteritems(items)
        )


def get_plugins():
    '''
    Get the list of installed plugins.
    '''

    module_filenames = glob(path.join(path.dirname(__file__), 'plugins', '*.py'))

    module_names = [
        path.basename(name)[:-3]
        for name in module_filenames
        if not name.endswith('__init__.py')
    ]

    # Import all the modules to populate PLUGINS
    for name in module_names:
        import_module('canaryd.plugins.{0}'.format(name))

    return PLUGINS


def get_plugin_names():
    return six.iterkeys(NAME_TO_PLUGIN)


def get_plugin_by_name(plugin_name):
    return NAME_TO_PLUGIN.get(plugin_name)


def prepare_plugin(plugin):
    logger.debug('Preparing plugin: {0}'.format(plugin))

    try:
        data = check_output(
            plugin.prepare_command or plugin.command,
            stderr=PIPE,
            shell=True,
        )

    except (CalledProcessError, OSError) as e:
        logger.warning('Error preparing plugin: {0}: {1}'.format(
            plugin.name, e,
        ))
        return False

    if hasattr(plugin, 'prepare'):
        plugin.prepare(data)

    return True


def get_and_prepare_working_plugins():
    all_plugins = get_plugins()
    working_plugins = []

    for plugin in all_plugins:
        status = prepare_plugin(plugin)

        if status:
            working_plugins.append(plugin)

    logger.info('Loaded plugins: {0}'.format(', '.join([
        plugin.__class__.__name__
        for plugin in working_plugins
    ])))

    return working_plugins


def get_plugin_state(plugin):
    '''
    Gets state output from a single plugin.
    '''

    if not plugin.command:
        return False, AttributeError('Invalid command')

    logger.debug('Running plugin: {0}'.format(plugin))

    try:
        data = check_output(
            plugin.command,
            stderr=PIPE,
            shell=True,
        )

    except Exception as e:
        logger.debug('Error running plugin: {0}: {1}', plugin, e)
        print_exception()
        return False, e

    try:
        state = plugin.parse(data)

    except Exception as e:
        logger.debug('Error parsing plugin: {0}: {1}'.format(plugin, e))
        print_exception()
        return False, e

    if not isinstance(state, (dict, list)):
        return False, TypeError('Invalid state type: {0}'.format(state))

    return True, state


def get_plugin_states(plugins):
    '''
    Gets state output from multiple plugins.
    '''

    return [
        (plugin, get_plugin_state(plugin))
        for plugin in plugins
    ]



import socket

from json import dumps as json_dumps, JSONEncoder
from time import sleep

from canaryd.log import logger
from canaryd.packages import requests
from canaryd.settings import CanarydSettings
from canaryd.version import __version__

SESSION = None


class CanaryJSONEncoder(JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'serialise'):
            return obj.serialise()

        if hasattr(obj, 'isoformat'):
            return obj.isoformat()

        return JSONEncoder.default(self, obj)


def backoff(function, *args, **kwargs):
    data = None
    interval = 0

    error_message = kwargs.get('error_message', 'API error')

    while data is None:
        try:
            return function(*args)

        except ApiError as e:
            if interval <= 290:
                interval += 10

            e.log()

        logger.critical('{0}, retrying in {1}s'.format(error_message, interval))
        sleep(interval)


class ApiError(Exception):
    '''
    Generic exception for problems with the API.
    '''

    def __init__(self, status_code, name, message=None, content=None):
        self.status_code = status_code
        self.name = name
        self.message = message
        self.content = content

    def log(self):
        logger.critical(
            '{0}: {1}{2}'.format(
                self.status_code,
                self.name,
                '({0})'.format(self.message) if self.message else '',
            ),
        )

        if self.content:
            logger.debug('Response data: {0}'.format(self.content))


def get_session():
    global SESSION

    if not SESSION:
        SESSION = requests.Session()

    return SESSION


def make_states_dict(states):
    states_dict = {}

    for plugin, (status, state) in states:
        if status:
            states_dict[plugin.name] = plugin.serialise_state(state)

    return states_dict


def make_changes_dict(states):
    states_dict = {}
    for plugin, (status, changes) in states:
        if status:
            states_dict[plugin.name] = changes

    return states_dict


def make_api_request(
    method, endpoint,
    api_key=None, settings=None, json=None, **kwargs
):
    settings = settings or CanarydSettings()

    api_key = api_key or settings.api_key

    if json:
        kwargs['data'] = json_dumps(json, cls=CanaryJSONEncoder)
        kwargs['headers'] = {
            'Content-Type': 'application/json',
        }

    try:
        response = method(
            '{0}/v{1}/{2}'.format(
                settings.api_base,
                settings.api_version,
                endpoint,
            ),
            auth=('api', api_key),
            **kwargs
        )
    except requests.ConnectionError as e:
        raise ApiError(
            0,
            'Could not connect: {0}'.format(e),
        )

    # Try to get some response JSON data
    response_data = {}

    try:
        response_data = response.json()

    except ValueError:
        raise ApiError(
            response.status_code,
            'Invalid JSON response',
            content=response.content,
        )

    # Capture and re-raise any HTTP/API errors
    try:
        response.raise_for_status()

    except requests.HTTPError as e:
        raise ApiError(
            response.status_code,
            response_data.get('error_name', 'Unknown'),
            message=response_data.get('error_message', e),
            content=response_data,
        )

    return response_data


def ping(settings):
    response_data = make_api_request(
        get_session().get,
        'server/{0}/ping'.format(settings.server_id),
        settings=settings,
    )

    return response_data.get('ping') == 'pong'


def upload_states_return_settings(url, states, settings, json=None):
    json = json or states

    if json is not states:
        json['states'] = states

    response_data = make_api_request(
        get_session().post, url,
        settings=settings,
        json=json,
    )

    return response_data['settings']


def sync_states(states, settings):
    '''
    Uploads a full state to api.servicecanary.com and returns any data sent back
    from the server (settings).
    '''

    return upload_states_return_settings(
        'server/{0}/sync'.format(settings.server_id),
        make_states_dict(states),
        settings,
        json={
            'hostname': socket.gethostname(),
            'canaryd_version': __version__,
        },
    )


def upload_state_changes(states, settings):
    '''
    Uploads partial state to api.servicecanary.com.
    '''

    return upload_states_return_settings(
        'server/{0}/state'.format(settings.server_id),
        make_changes_dict(states),
        settings,
    )


def register(key):
    '''
    Register this server on api.servicecanary.com and return a server id.
    '''

    response_data = make_api_request(
        get_session().post,
        'servers',
        api_key=key,
        json={
            'hostname': socket.gethostname(),
            'canaryd_version': __version__,
        },
    )

    return response_data['server_id']


def signup(email):
    '''
    Signup for api.servicecanary.com and return the access key for that user.
    '''

    response_data = make_api_request(
        get_session().post,
        'signup',
        json={
            'email': email,
        },
    )

    if 'api_key' in response_data:
        return True, response_data['api_key']

    return False, response_data['info']



# cansryd
# File: canaryd/canaryct/__init__.py
# Desc: canaryctl functions

from __future__ import print_function

import json

from os import path, system

from canaryd.packages import click

from canaryd import remote

from canaryd.exceptions import CanarydError
from canaryd.log import logger, setup_logging
from canaryd.plugin import (
    get_plugin_state,
    get_plugins,
    get_plugin_by_name,
    prepare_plugin,
)
from canaryd.remote import CanaryJSONEncoder
from canaryd.settings import (
    CanarydSettings,
    get_config_directory,
    get_config_file,
    get_settings,
    write_settings_to_config,
)
from canaryd.version import __version__

from .check_root import check_root
from .install_service import install_service


# Parse arguments
@click.group()
@click.option('-v', '--verbose', count=True)
def main(verbose=0):
    '''
    canaryd control.
    '''

    # For canaryctl we want warnings to show, so always bump verbosity
    verbose += 1
    setup_logging(verbose)


@main.command()
@click.option('--start', is_flag=True, default=False)
@click.argument('key', required=False)
@click.pass_context
def init(ctx, start, key):
    '''
    Create the canaryd service and start it.

    This command will attempt to register if the config file is not found.
    '''

    check_root('Not root user, cannot create services.', exit=True)

    config_file = get_config_file()

    # Register if the config file cannot be found
    if not path.exists(config_file):
        did_register = ctx.invoke(register, key=key)

        if not did_register:
            raise CanarydError('Failed to register')

    # Install the service
    click.echo('--> Installing canaryd service')
    start_command = install_service()

    if (
        start or
        click.confirm(
            'Start canaryd service ({0})?'.format(start_command),
            default=True,
        )
    ):
        system(start_command)


@main.command()
@click.argument('key', required=False)
def register(key):
    '''
    Register this server on Service Canary.

    If no api key is provided, you can sign up instantly.
    '''

    check_root('''
Not root user, using config directory:
{0}
It is recommended to run canaryd as root as
plugins that require privileges (eg iptables)
will not function properly.
'''.strip().format(get_config_directory()))

    config_file = get_config_file()

    if not key:
        click.echo('--> No key provided or set for this instance')
        click.echo('--> To sign up, input your email address, or blank to skip')
        email_or_blank = raw_input('Email or blank: ')

        if '@' not in email_or_blank:
            return False

        # Signup and get the key
        did_signup, key_or_message = remote.signup(email_or_blank)

        if did_signup:
            key = key_or_message
            click.echo('--> You are now signed up for servicecanary.com.')
            click.echo('--> Check your email for a login link to view updates.')

        else:
            click.echo(click.style(key_or_message, 'blue'))
            return False

    # Register the server
    server_id = remote.register(key=key)

    # Create our settings
    settings = CanarydSettings(api_key=key, server_id=server_id)

    # Write the settings to the config file
    write_settings_to_config(settings)

    click.echo('--> {0} written'.format(config_file))

    return True


@main.command()
@click.argument('plugin')
def state(plugin):
    '''
    Get state for a single plugin.
    '''

    target_plugin = get_plugin_by_name(plugin)

    if not target_plugin:
        raise TypeError('Invalid plugin: {0}'.format(plugin))

    prepare_plugin(target_plugin)

    click.echo('State for {0}:'.format(plugin))

    status, data = get_plugin_state(target_plugin)

    if status:
        print(json.dumps(
            target_plugin.serialise_state(data),
            cls=CanaryJSONEncoder,
            indent=4,
        ))
    else:
        logger.critical((
            'Unexpected exception while getting {0} state: '
            '{1}({2})'
        ).format(plugin.name, data.__class__.__name__, data))


@main.command()
def plugins():
    '''
    List all plugins.
    '''

    click.echo('--> Available plugins: {0}'.format(', '.join(
        plugin.name for plugin in get_plugins()
    )))


@main.command()
def version():
    '''
    Print the canaryd version and config location.
    '''

    click.echo('canaryd: v{0}'.format(__version__))
    click.echo('config: {0}'.format(get_config_file()))


@main.command()
def ping():
    '''
    Ping servicecanary.com.
    '''

    config_file = get_config_file()

    if not path.exists(config_file):
        click.echo((
            'No config file ({0}) exists, '
            'please run `canaryctl init`'
        ).format(config_file))
        return

    settings = get_settings(config_file)

    # Ping the API
    remote.ping(settings)

    click.echo('OK!')



from canaryd.packages import six
from canaryd.packages.dmidecode import parse_dmi

from canaryd.plugin import Plugin


def slugify(value):
    return value.lower().replace(' ', '_')


class Hardware(Plugin):
    spec = ('key', {
        'type': str,
        'serial_number': str,
        'version': str,
        'meta': dict,
    })

    command = 'dmidecode'

    @staticmethod
    def parse(output):
        dmi_data = parse_dmi(output)

        data = []

        for key, meta in dmi_data:
            # Make _title -> title
            meta['title'] = meta.pop('_title')

            meta = dict(
                (slugify(k), value)
                for k, value in six.iteritems(meta)
            )

            item = {}
            for k in ('type', 'serial_number', 'version'):
                if k in meta:
                    item[k] = meta.pop(k)

            item['meta'] = meta

            data.append((slugify(key), item))

        return dict(data)



class ServiceMixin(object):
    @staticmethod
    def event_message(type_, key, data_changes):
        if type_ != 'updated':
            return

        if 'pid' in data_changes:
            return 'Service restarted: {0}'.format(key)

        if 'running' in data_changes:
            running = data_changes['running'][1]

            if running:
                return 'Service started: {0}'.format(key)
            else:
                return 'Service stopped: {0}'.format(key)



import re
import socket

from datetime import datetime, timedelta

from canaryd.plugin import Plugin

UPTIME_REGEX = re.compile((
    'up\s+(.*?),\s+([0-9]+) '
    'users?,\s+load averages?: '
    '([0-9]+\.[0-9][0-9]),?\s+([0-9]+\.[0-9][0-9]),?\s+([0-9]+\.[0-9][0-9])'
))


def ensure_datetime(datetime_or_string):
    if isinstance(datetime_or_string, datetime):
        return datetime_or_string

    return datetime.strptime(datetime_or_string, '%Y-%m-%dT%H:%M:%S')


class Meta(Plugin):
    spec = ('key', {
        'value': None,
    })

    command = 'uptime'

    @staticmethod
    def parse(output):
        data = [
            ('hostname', socket.gethostname()),
        ]

        for line in output.splitlines():
            line = line.strip()
            matches = re.search(UPTIME_REGEX, line)

            if matches:
                duration, users, av1, av5, av15 = matches.groups()
                data.append(('load_average', [float(av1), float(av5), float(av15)]))
                data.append(('users', int(users)))

                days = 0
                hours = 0
                mins = 0

                if 'day' in duration:
                    match = re.search('([0-9]+)\s+day', duration)
                    days = int(match.group(1))

                if ':' in duration:
                    match = re.search('([0-9]+):([0-9]+)', duration)
                    hours = int(match.group(1))
                    mins = int(match.group(2))

                if 'min' in duration:
                    match = re.search('([0-9]+)\s+min', duration)
                    mins = int(match.group(1))

                up_since = datetime.utcnow() - timedelta(
                    days=days, hours=hours, minutes=mins,
                )

                up_since = up_since.replace(
                    second=0,
                    microsecond=0,
                )

                data.append(('up_since', up_since))

        # Nest each value in a dict
        return dict(
            (key, {
                'value': value,
            })
            for key, value in data
        )

    @staticmethod
    def is_change(key, previous_item, item):
        if key == 'hostname':
            return True

        if key == 'up_since':
            previous_up = ensure_datetime(previous_item['value'])
            up = ensure_datetime(item['value'])

            if (
                previous_up.day == up.day
                and previous_up.hour == up.hour
                # Uptime is 1+ minutes less than previous
                and (previous_up.minute + 1) < up.minute
            ):
                return True

        return False

    @staticmethod
    def event_message(type_, key, data_changes):
        if type_ != 'updated':
            return

        if key == 'up_since':
            return 'Server rebooted'

        if key == 'hostname':
            return 'Hostname changed'



# canaryd
# File: canaryd/__main__.py
# Desc: entry point for canaryd

import logging

from time import time

from canaryd.packages import click  # noqa

from canaryd.daemon import run_daemon
from canaryd.log import logger, setup_logging, setup_file_logging
from canaryd.plugin import get_and_prepare_working_plugins, get_plugin_states
from canaryd.remote import backoff, ping, sync_states
from canaryd.settings import get_config_file, get_settings
from canaryd.version import __version__


@click.command()
@click.option('-v', '--verbose', count=True)
def main(verbose=0):
    '''
    Run the canaryd daemon.
    '''

    setup_logging(verbose)
    logger.info('Starting canaryd v{0}'.format(__version__))

    config_file = get_config_file()

    # Load the settings, using our config file if provided
    settings = get_settings(config_file)

    # Setup any log file
    if settings.log_file:
        setup_file_logging(settings.log_file)

    # Debug setting override
    if settings.debug == 'true':
        logger.setLevel(logging.DEBUG)

    if not settings.api_key or not settings.server_id:
        logger.critical('Missing api_key and/or server_id in config file')
        return

    # Initial ping for API presence
    backoff(
        ping, settings,
        error_message='Could not ping',
    )

    # Load the plugin list
    plugins = get_and_prepare_working_plugins()

    # Get the initial state
    logger.info('Getting initial state...')
    start_time = time()
    states = get_plugin_states(plugins)

    # Sync this state and get settings
    logger.info('Syncing initial state...')

    remote_settings = backoff(
        sync_states, states, settings,
        error_message='Could not sync state',
    )

    # Update settings w/remote ones
    settings.update(remote_settings)

    # Run the loop
    logger.info('Starting daemon loop...')

    # Make previous states dict
    previous_states = dict(
        (plugin, status_data[1])
        for plugin, status_data in states
        if status_data[0]
    )

    try:
        run_daemon(plugins, previous_states, settings, start_time=start_time)

    # This is OK!
    except KeyboardInterrupt:
        logger.info('Exiting canaryd...')


main()



# canaryd
# File: canaryd/daemon.py
# Desc: the canaryd daemon

from time import time, sleep

from canaryd.diff import get_state_diff
from canaryd.log import logger
from canaryd.plugin import get_plugin_states
from canaryd.remote import upload_state_changes, ApiError


def _sleep_until_interval(start, interval):
    time_taken = time() - start

    if time_taken < interval:
        sleep(interval - time_taken)


def _daemon_loop(plugins, previous_states, settings):
    states = get_plugin_states(plugins)
    state_changes = []

    for plugin, status_data in states:
        status, data = status_data

        if status:
            state_diff = get_state_diff(
                plugin, data,
                previous_states.get(plugin, {}),
            )

            state_changes.append((plugin, (status, state_diff)))

            previous_states[plugin] = data
        else:
            logger.critical((
                'Unexpected exception while getting {0} state: '
                '{1}({2})'
            ).format(plugin.name, data.__class__.__name__, data))
            continue

    try:
        settings_changes = upload_state_changes(state_changes, settings)

        if settings_changes:
            settings.update(settings_changes)

    except ApiError as e:
        e.log()


def run_daemon(plugins, previous_states, settings, start_time=None):
    if start_time:
        _sleep_until_interval(
            start_time, settings.collect_interval_s,
        )

    while True:
        start = time()

        _daemon_loop(plugins, previous_states, settings)

        _sleep_until_interval(
            start, settings.collect_interval_s,
        )



from canaryd.plugin import Plugin

from .mixins.service import ServiceMixin

# We ignore these as they regularly get deleted/added as part of normal OSX
# lifecycle - and as such any events generated are not of use.
IGNORE_NAMES = ('oneshot', 'mdworker', 'mbfloagent')


class Launchctl(Plugin, ServiceMixin):
    spec = ('service', {
        'running': bool,
        'pid': int,
    })

    command = 'launchctl list'

    @staticmethod
    def parse(output):
        state = {}

        for line in output.splitlines():
            if any(name in line for name in IGNORE_NAMES):
                continue

            bits = line.split()

            if not bits or bits[0] == 'PID':
                continue

            name = bits[2]

            # If the last "bit" is just a number, it's the PID, so we strip it
            name_bits = name.split('.')
            if name_bits[-1].isdigit():
                name = '.'.join(name_bits[:-1])

            data = {}

            try:
                data['pid'] = int(bits[0])
            except ValueError:
                pass

            data['running'] = 'pid' in data

            state[name] = data

        return state



import re

from subprocess import CalledProcessError

from canaryd.packages.check_output import check_output

from canaryd.plugin import Plugin

from .mixins.service import ServiceMixin

SYSTEMD_REGEX = r'^([a-z\-]+)\.service\s+[a-z\-]+\s+[a-z]+\s+([a-z]+)'


class Systemd(Plugin, ServiceMixin):
    spec = ('service', {
        'running': bool,
        'pid': int,
    })

    command = 'systemctl -alt service list-units'

    @staticmethod
    def parse(output):
        services = {}

        for line in output.splitlines():
            line = line.strip()
            matches = re.match(SYSTEMD_REGEX, line)
            if matches:
                name = matches.group(1)

                pid = None

                try:
                    pid = check_output(
                        'systemctl status {0}.service | grep "Main PID:"'.format(name),
                        shell=True,
                    )
                    pid = pid.split()

                    for bit in pid:
                        try:
                            pid = int(bit)
                            break
                        except ValueError:
                            pass

                except (CalledProcessError, OSError):
                    pass

                services[name] = {
                    'running': matches.group(2) == 'running',
                    'pid': pid,
                }

        return services



# canaryd
# File: canaryd/settings.py
# Desc: settings for canaryd/canaryctl

import platform

from ConfigParser import RawConfigParser, Error as ConfigParserError
from os import environ, geteuid, makedirs, path

from canaryd.packages import click, six

from canaryd.exceptions import ConfigError
from canaryd.log import logger

API_BASE = environ.get(
    'API_BASE',
    'https://api.servicecanary.com',
)

API_VERSION = int(environ.get(
    'API_VERSION',
    1,
))

API_KEY = environ.get(
    'API_KEY',
    None,
)

SERVER_ID = environ.get(
    'SERVER_ID',
    None,
)


class CanarydSettings(object):
    api_base = API_BASE
    api_version = API_VERSION

    log_file = None
    debug = False

    collect_interval_s = 60

    # API access key
    api_key = API_KEY

    # ID of (hopefully this) server attached to this key
    server_id = SERVER_ID

    def __init__(self, **kwargs):
        self.update(kwargs)

    def update(self, data):
        for key, value in six.iteritems(data):
            setattr(self, key, value)

        logger.debug('Settings updated: {0}'.format(data))


def get_config_directory():
    # If we're non-root, just use the users config dir (~/.config, etc)
    if geteuid() > 0:
        config_directory = click.get_app_dir('canaryd')

        return config_directory

    # If we're root on OSX
    if platform.system() == 'Darwin':
        return path.join('/', 'Library', 'Application Support', 'canaryd')

    # Elsewhere on *nix (no Windows support _yet_)
    return path.join('/', 'etc', 'canaryd')


def get_config_file():
    return path.join(get_config_directory(), 'canaryd.conf')


def get_settings(config_file):
    '''
    Load the config from the filesystem if provided, with defaults.
    '''

    settings = CanarydSettings()

    parser = RawConfigParser()

    try:
        parser.read(config_file)
        canaryd_config_items = parser.items('canaryd')

    except ConfigParserError as e:
        logger.critical('Error in config file ({0}): {1}'.format(config_file, e.message))
        raise ConfigError('Config file error')

    for key, value in canaryd_config_items:
        if hasattr(settings, key):
            setattr(settings, key, value)

    logger.info('Loaded settings file: {0}'.format(config_file))

    return settings


def write_settings_to_config(settings):
    '''
    Write a config file from settings.
    '''

    # Ensure the config directory exists
    config_directory = get_config_directory()

    if not path.exists(config_directory):
        makedirs(config_directory)

    # Generate the config
    config = RawConfigParser()
    config.add_section('canaryd')

    for key, value in six.iteritems(settings.__dict__):
        config.set('canaryd', key, value)

    # Write the config file
    config_file = get_config_file()

    with open(config_file, 'wb') as f:
        config.write(f)



from os import path
from pkgutil import get_data
from subprocess import CalledProcessError, PIPE

from canaryd.packages import click
from canaryd.packages.check_output import check_output


def which(command):
    try:
        return check_output(
            ('which', command),
            stderr=PIPE,
        ).strip()

    except (CalledProcessError, OSError):
        pass


def write_script(script_path, script):
    canaryd_location = which('canaryd')

    script = script.replace('CANARYD_LOCATION', canaryd_location)

    # Write the init system script
    with open(script_path, 'w') as file:
        file.write(script)


def install_service():
    script = None
    script_path = None
    start_command = None

    # OSX/Darwin
    if which('launchctl'):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.plist'))
        script_path = path.join('/', 'Library', 'LaunchDaemons', 'canaryd.plist')
        start_command = 'launchctl start canaryd'

    # Systemd
    elif which('systemctl'):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.service'))
        script_path = path.join('/', 'etc', 'systemd', 'system', 'canaryd.service')
        start_command = 'systemctl start canaryd.service'

    # Upstart
    elif which('initctl'):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.conf'))
        script_path = path.join('/', 'etc', 'init', 'canaryd.conf')
        start_command = 'start canaryd'

    # Init.d
    elif path.exists(path.join('/', 'etc', 'init.d')):
        script = get_data('canaryd', path.join('init_scripts', 'canaryd.sh'))
        script_path = path.join('/', 'etc', 'init.d', 'canaryd')
        start_command = 'chmod +x /etc/init.d/canaryd && /etc/init.d/canaryd start'

    if any(item is None for item in (script, script_path, start_command)):
        pass

    write_script(script_path, script)
    click.echo('--> {0} written'.format(script_path))

    return start_command



import re

from canaryd.plugin import Plugin

from .mixins.service import ServiceMixin

SERVICE_REGEX = r'([a-zA-Z0-9\-]+)=([0-9]+)=([0-9]+)?'


class Initd(Plugin, ServiceMixin):
    spec = ('service', {
        'running': bool,
        'pid': int,
    })

    command = '''
        for SERVICE in `ls /etc/init.d/`; do
            _=`cat /etc/init.d/$SERVICE | grep status`
            if [ "$?" = "0" ]; then
                PID=` \
                    ps --ppid 1 -o 'tty,pid,comm' \
                    | grep ^?.*$SERVICE \
                    | head -n 1 \
                    | sed -n -e 's/?\s\+\([0-9]\+\)\s\+.*/\\1/p' \
                `
                STATUS=`/etc/init.d/$SERVICE status`
                echo "$SERVICE=$?=$PID"
            fi
        done
    '''

    @staticmethod
    def parse(output):
        services = {}

        for line in output.splitlines():
            matches = re.match(SERVICE_REGEX, line)
            if matches:
                status = int(matches.group(2))
                pid = matches.group(3)

                if pid:
                    pid = int(pid)

                # Exit code 0 = OK/running
                if status == 0:
                    status = True

                # Exit codes 1-3 = DOWN/not running
                elif status < 4:
                    status = False

                # Exit codes 4+ = unknown
                else:
                    status = None

                services[matches.group(1)] = {
                    'running': status or isinstance(pid, int),
                    'pid': pid,
                }

        return services



# canaryd
# File: canaryd/ctl/__main__.py
# Desc: entry point for canaryctl

from __future__ import print_function

from canaryd.log import logger
from canaryd.remote import ApiError

from . import main


try:
    main()

except ApiError as e:
    logger.critical('API {0} error: {1}({2})'.format(
        e.status_code,
        e.name,
        e.message,
    ))



import re

from canaryd.plugin import Plugin

USER_REGEX = r'^uid=[0-9]+\(([a-zA-Z0-9_\.\-]+)\) gid=[0-9]+\(([a-zA-Z0-9_\.\-]+)\) groups=([a-zA-Z0-9_\.\-,\(\)\s]+) (.*)$'  # noqa
GROUP_REGEX = r'^[0-9]+\(([a-zA-Z0-9_\.\-]+)\)$'


class Users(Plugin):
    spec = ('user', {
        'group': str,
        'groups': set((str,)),
        'home': str,
        'shell': str,
    })

    command = '''
        for i in `cat /etc/passwd | cut -d: -f1`; do
            ID=`id $i`
            META=`cat /etc/passwd | grep ^$i: | cut -d: -f6-7`
            echo "$ID $META"
        done
    '''

    @staticmethod
    def parse(output):
        users = {}

        for line in output.splitlines():
                matches = re.match(USER_REGEX, line)

                if matches:
                    # Parse out the home/shell
                    home_shell = matches.group(4)
                    home = shell = None

                    # /blah: is just a home
                    if home_shell.endswith(':'):
                        home = home_shell[:1]

                    # :/blah is just a shell
                    elif home_shell.startswith(':'):
                        shell = home_shell[1:]

                    # Both home & shell
                    elif ':' in home_shell:
                        home, shell = home_shell.split(':')

                    # Main user group
                    group = matches.group(2)

                    # Parse the groups
                    groups = []
                    for group_matches in matches.group(3).split(','):
                        name = re.match(GROUP_REGEX, group_matches.strip())
                        if name:
                            name = name.group(1)
                        else:
                            continue

                        # We only want secondary groups here
                        if name != group:
                            groups.append(name)

                    groups = set(groups)

                    users[matches.group(1)] = {
                        'group': group,
                        'groups': groups,
                        'home': home,
                        'shell': shell,
                    }

        return users



from subprocess import CalledProcessError, PIPE, Popen


def check_output(*popenargs, **kwargs):
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, _ = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get('args')
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd)
    return output



import re

from canaryd.plugin import Plugin

from .mixins.service import ServiceMixin

UPSTART_REGEX = r'^([a-z\-]+) [a-z]+\/([a-z]+),?\s?(process)?\s?([0-9]+)?'


class Upstart(Plugin, ServiceMixin):
    spec = ('service', {
        'running': bool,
        'pid': int,
    })

    command = 'initctl list'

    @staticmethod
    def parse(output):
        services = {}

        for line in output.splitlines():
            matches = re.match(UPSTART_REGEX, line)
            if matches:
                pid = matches.group(4)

                if pid:
                    pid = int(pid)

                services[matches.group(1)] = {
                    'running': matches.group(2) == 'running',
                    'pid': pid,
                }

        return services
